#include "../../../internal-include/CD/cd_internal.hpp"
#include <PSX/File/Processor/cd_file_processor.hpp>
#include <stdio.h>

namespace JabyEngine {
    static constexpr auto NormalCircularBufferSize = 5;

    void CDFileProcessor :: start_cur_job() {
        using CD::internal::FileInfo;
        using CD::internal::SectorBufferAllocator;
            const auto configurate_for = [this](const CDFile& file) -> SectorBufferAllocator {
                static const auto circular_buffer_callback = [](void* ctx) -> CD_IO::DataSector* {
                    CDFileProcessor &self = *reinterpret_cast<CDFileProcessor*>(ctx);
                    return self.circular_buffer.allocate();
                };

                switch(file.type) {
                    case CDFileType::SimpleTIM:
                        printf("CDFileProcessor: SimpleTIM not supported yet\n");
                        return SectorBufferAllocator::invalid();

                    case CDFileType::CopyTo:
                        this->circular_buffer.setup(reinterpret_cast<CD_IO::DataSector*>(file.payload.copy_to.dst), 512);
                        this->lz4_decomp.disable();

                        this->file_state = FileProcessor::create(this->tmp_area, Nothing());
                        return SectorBufferAllocator::create(this, circular_buffer_callback);

                    default:
                        return SectorBufferAllocator::invalid();
                }            
            };

        const auto& cur_job = *this->jobs.files;
        const auto& cur_lba = this->lba[cur_job.rel_lba_idx];
        const auto  cfg     = configurate_for(cur_job);

        CD::internal::read_file(FileInfo::from(cur_lba), cfg);

        printf(">>> CD needs to load LBA: %i -> %i\n", cur_lba.lba, cur_lba.size_words);
    }

    bool CDFileProcessor :: process_data() {
        while(this->circular_buffer.has_data()) {
            ArrayRange<const uint8_t> cur_sector(reinterpret_cast<uint8_t*>(this->circular_buffer.get_next()->data), CD_IO::DataSector::SizeBytes);
            
            //                                                       v We can not know if there will be more data or not - but it also doesn't matter much for us
            const auto result = this->lz4_decomp.process(cur_sector, false);
            this->circular_buffer.pop();

            if(result) {
                // Process the data in the tmp_area
                if(this->file_state.process(result.bytes_ready/sizeof(uint32_t)) == Progress::Error) {
                    return false;
                }
            }

            else {
                return false;
            }
        }
        return true;
    }

    void CDFileProcessor :: setup(const volatile AutoLBAEntry* lba, JobArray jobs, uint32_t* tmp_area) {
        this->lba      = const_cast<const AutoLBAEntry*>(lba);
        this->jobs     = jobs;

        // Setsup the circular buffer and determines where to place the temp area in
        const_cast<uint32_t*&>(this->tmp_area) = reinterpret_cast<uint32_t*>(this->circular_buffer.setup(reinterpret_cast<CD_IO::DataSector*>(tmp_area), NormalCircularBufferSize));
        CDFileProcessor::start_cur_job();
    }

    Progress CDFileProcessor :: process() {
        CDFileProcessor::process_data();
        switch(CD::internal::read_current_state()) {
            case CD::internal::State::Done:
                /* 
                    We are done now!
                    The user decides if he wants the next value
                */                
                return Progress::Done;

            case CD::internal::State::BufferFull:
                /* We processd data and unpause the CD drive */
                return CD::internal::continue_reading() ? Progress::InProgress : Progress::Error;

            case CD::internal::State::Reading:
                return Progress::InProgress;

            case CD::internal::State::Error:
            default:
                /* Error for real */
                return Progress::Error;
        }
    }
}